import os

output_functions = {}

for net in netlist.nets:
    if net.name.startswith('c('):
        i = int(net.name.replace('c(','').replace(')',''))
        output_functions[i] = hal_py.NetlistUtils.get_subgraph_function(net, netlist.gates)


# get multibit boolean functions
output_c = output_functions[0]
for i in range(1, 32):
    new_size = output_functions[i].get_size() + output_c.get_size()
    output_c = hal_py.BooleanFunction.Concat(output_functions[i], output_c, new_size)



# substitute variable names with their original input
# important: get rid of brackets, due to smt2 syntax
for net in netlist.nets:
    output_c = output_c.substitute("net_" + str(net.get_id()), net.get_name().replace("(", "").replace(")",""))



# check if adder
var_a = hal_py.BooleanFunction.Var("a0", 1)
var_b = hal_py.BooleanFunction.Var("b0", 1)
for i in range(1,32):
    new_size = 1 + var_a.get_size()
    var_a = hal_py.BooleanFunction.Concat(hal_py.BooleanFunction.Var("a" +str(i), 1), var_a, new_size)
    var_b = hal_py.BooleanFunction.Concat(hal_py.BooleanFunction.Var("b" +str(i), 1), var_b, new_size)

var_p = hal_py.BooleanFunction.Var("password0", 1)
for i in range(1,48):
    new_size = 1 + var_p.get_size()
    var_p = hal_py.BooleanFunction.Concat(hal_py.BooleanFunction.Var("password" +str(i), 1), var_p, new_size)



#build constraints
const_one = hal_py.BooleanFunction.Const(hal_py.BooleanFunction.Value.ONE)
const_zero = hal_py.BooleanFunction.Const(hal_py.BooleanFunction.Value.ZERO)
def const_rand():
    r = os.urandom(1)[0]&1
    return const_one if r==1 else const_zero

bf_adder = hal_py.BooleanFunction.Eq(hal_py.BooleanFunction.Add(var_a, var_b, 32), output_c, 1)

constrs = []

for run_no in range(128):
    print(run_no)
    dic = {}
    for i in range(32):
        dic[f"a{i}"] = const_rand()
        dic[f"b{i}"] = const_rand()
    special_case = bf_adder.substitute(dic)
    #print(str(special_case))
    constrs.append(hal_py.SMT.Constraint(special_case))


# This uses SMT solvers to find a set of inputs that make the trigger function evaluate to 1.
solver = hal_py.SMT.Solver(constrs)
res = solver.query(hal_py.SMT.QueryConfig().with_solver(hal_py.SMT.SolverType.Z3).with_local_solver().with_model_generation().with_timeout(1000))
print(res.type)
p_res = res.model.evaluate(var_p)
print('password:', p_res)
print(int(str(p_res),2).to_bytes(48//8))

